"""
Crochet: Use Twisted Anywhere!
"""

from __future__ import absolute_import

import sys

from twisted.python.log import startLoggingWithObserver
from twisted.python.runtime import platform

from ._shutdown import _watchdog, register
from ._eventloop import (
    EventualResult, EventLoop, _store, ReactorStopped
)
from ._eventloop import TimeoutError  # pylint: disable=redefined-builtin
from ._version import get_versions

if platform.type == "posix":
    try:
        from twisted.internet.process import reapAllProcesses
    except (SyntaxError, ImportError):
        if sys.version_info < (3, 3, 0):
            raise
        else:
            # Process support is still not ported to Python 3 on some versions
            # of Twisted.
            def reapAllProcesses(): pass
else:
    # waitpid() is only necessary on POSIX:
    def reapAllProcesses(): pass


__version__ = get_versions()['version']
del get_versions


def _importReactor():
    from twisted.internet import reactor
    return reactor


_main = EventLoop(
    _importReactor, register, startLoggingWithObserver, _watchdog,
    reapAllProcesses)
setup = _main.setup
no_setup = _main.no_setup
run_in_reactor = _main.run_in_reactor
wait_for = _main.wait_for
retrieve_result = _store.retrieve

# Backwards compatibility with 0.5.0:
in_reactor = _main.in_reactor
DeferredResult = EventualResult

# Backwards compatibility with 1.1.0 and earlier:
wait_for_reactor = _main.wait_for_reactor

__all__ = [
    "setup",
    "run_in_reactor",
    "EventualResult",
    "TimeoutError",
    "retrieve_result",
    "no_setup",
    "wait_for",
    "ReactorStopped",
    "__version__",
]



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import versioneer


def read(path):
    """
    Read the contents of a file.
    """
    with open(path) as f:
        return f.read()


setup(
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    name='crochet',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Use Twisted anywhere!",
    install_requires=[
        "Twisted>=16.0",
        "wrapt",
    ],
    keywords="twisted threading",
    license="MIT",
    packages=["crochet", "crochet.tests"],
    url="https://github.com/itamarst/crochet",
    maintainer='Itamar Turner-Trauring',
    maintainer_email='itamar@itamarst.org',
    long_description=read('README.rst') + '\n' + read('docs/news.rst'),
)



"""Tests for the logging bridge."""

from __future__ import absolute_import

from twisted.trial.unittest import SynchronousTestCase
import threading

from twisted.python import threadable

from .._eventloop import ThreadLogObserver


class ThreadLogObserverTest(SynchronousTestCase):
    """
    Tests for ThreadLogObserver.

    We use Twisted's SyncTestCase to ensure that unhandled logged errors get
    reported as errors, in particular for test_error.
    """

    def test_stop(self):
        """
        ThreadLogObserver.stop() stops the thread started in __init__.
        """
        threadLog = ThreadLogObserver(None)
        self.assertTrue(threadLog._thread.is_alive())
        threadLog.stop()
        threadLog._thread.join()
        self.assertFalse(threadLog._thread.is_alive())

    def test_emit(self):
        """
        ThreadLogObserver.emit runs the wrapped observer's in its thread, with
        the given message.
        """
        messages = []

        def observer(msg):
            messages.append((threading.current_thread().ident, msg))

        threadLog = ThreadLogObserver(observer)
        ident = threadLog._thread.ident
        msg1 = {}
        msg2 = {"a": "b"}
        threadLog(msg1)
        threadLog(msg2)
        threadLog.stop()
        # Wait for writing to finish:
        threadLog._thread.join()
        self.assertEqual(messages, [(ident, msg1), (ident, msg2)])

    def test_errors(self):
        """
        ThreadLogObserver.emit catches and silently drops exceptions from its
        observer.
        """
        messages = []
        counter = []

        def observer(msg):
            counter.append(1)
            if len(counter) == 2:
                raise RuntimeError("ono a bug")
            messages.append(msg)

        threadLog = ThreadLogObserver(observer)
        msg1 = {"m": "1"}
        msg2 = {"m": "2"}
        msg3 = {"m": "3"}
        threadLog(msg1)
        threadLog(msg2)
        threadLog(msg3)
        threadLog.stop()
        # Wait for writing to finish:
        threadLog._thread.join()
        self.assertEqual(messages, [msg1, msg3])

    def test_ioThreadUnchanged(self):
        """
        ThreadLogObserver does not change the Twisted I/O thread (which is
        supposed to match the thread the main reactor is running in.)
        """
        threadLog = ThreadLogObserver(None)
        threadLog.stop()
        threadLog._thread.join()
        self.assertIn(
            threadable.ioThread,
            # Either reactor was never run, or run in thread running
            # the tests:
            (None, threading.current_thread().ident))
